from mhelper import ansi, string_helper, AnsiStr, ansi_helper
from typing import List

import itertools

from sxsxml.sxs.base import SxsWriter
from sxsxml.support import XsProgress


class SxsAnsiWriter( SxsWriter ):
    """
    Writes ANSI from an SXS stream.
    """
    
    BAR = ansi.DIM
    RESET = "\0\0RE"
    
    
    def __init__( self, *args, **kwargs ):
        super().__init__( *args, **kwargs )
        
        self.__in_progress = False
        self.__last_out_eol = False
        self.__progress_not_finished = False
    
    
    def on_format_section_end( self ):
        return ""
    
    
    def on_format_text( self, x ):
        # If we are in a progress bar then we'll need to add a line break
        if self.__in_progress:
            self.__in_progress = False
            
            if self.__progress_not_finished:
                self.on_text( " »\n" )
            else:
                self.on_text( "\n" )
        
        self.__last_out_eol = x.endswith( "\n" )
        
        x = x.replace( "\n", "\n" + self.__section_prefix() )
        x = x.replace( self.RESET, ansi.RESET )
        return x
    
    
    def _exit_default( self ):
        text = self.pop()
        self.on_text( "{}{}{}".format( ansi.BLINK + str( self.element ) + ansi.BLINK_OFF, text, ansi.BLINK + "</>" + ansi.BLINK_OFF ) )
    
    
    def on_consume_eol( self ):
        if self.__in_progress or self.__last_out_eol:
            # Consume the EOL - return to line start and issue new prefix 
            self.on_text( "\r" )
            self.on_text( self.__section_prefix() )
        else:
            # Part way through a line - issue a new line
            self.on_text( "\n" )
    
    
    def __section_prefix( self ):
        if not self.section_depth:
            return ""
        
        return self.BAR + "│ " * self.section_depth + ansi.RESET
    
    
    def on_format_progress( self, p: XsProgress ):
        r = ("\r" + self.__section_prefix() + ansi.FORE_GREEN
             + "PROGRESS ({}) IS {} OF {}: '{}'".format( self.section_name,
                                                         p.value,
                                                         p.max,
                                                         p.text.ljust( 20, "." ) )
             + ansi.RESET)
        
        return r
    
    
    def on_formatted_progress( self, p: XsProgress ):
        self.ignore_next_eol()
        self.__in_progress = True
        self.__last_out_eol = False
        self.__progress_not_finished = p.value != p.max
    
    
    def on_format_heading( self, level, name ) -> str:
        if not name:
            return ""
        
        if level == 1:
            box = ansi.Box.DOUBLE
        elif level == 2:
            box = ansi.Box.BOLD
        elif level == 3:
            box = ansi.Box.SINGLE
        elif level == 3:
            box = ansi.Box.DOTTED
        else:
            box = None
        
        colour = ansi.RESET
        reset = ansi.RESET
        
        if box is not None:
            actual_name_ = name
            line_of_name = box.hr * len( name )
            above_name_ = " " + box.tl + box.ts + line_of_name + box.ts + box.tr
            centre_name = " " + box.ls + box.em + actual_name_ + box.em + box.rs
            below_name_ = " " + box.bl + box.bs + line_of_name + box.bs + box.br
            
            r = (self.BAR + "╭" + colour + string_helper.ljust( above_name_, 80, " " ) + reset + "\n"
                 + self.BAR + "│" + colour + string_helper.ljust( centre_name, 80, " " ) + reset + "\n"
                 + self.BAR + "│" + colour + string_helper.ljust( below_name_, 80, " " ) + reset)
        else:
            r = self.BAR + "╭" + colour + "*" * level + " " + name + reset
        
        return r
    
    
    def on_format_question( self, question, options ):
        question = ansi.FORE_BRIGHT_CYAN + question.strip() + ansi.RESET + "\n" + "\n".join( " > {}".format( ansi.FORE_BRIGHT_BLUE + x + ansi.RESET ) for x in options )
        question = ":: " + question.replace( "\n", "\n:: " )
        return question
    
    
    def on_format_table( self, rows ):
        # Note - we can't use a library because we need to preserve ansi escape sequences
        
        box = ansi.Box.BLANK
        col_lens = [max( len( AnsiStr( row[col].split( "\n" )[0] ) ) for row in rows ) for col in range( len( rows[0] ) )]
        r: List[str] = []
        x = [[("=" * l) for l in col_lens]]
        needs_l = False
        
        for row_index, row in enumerate( itertools.chain( x, [rows[0]], x, rows[1:], x ) ):
            row_lines = [list( ansi_helper.wrap( row[col], col_lens[col] ) ) for col in range( len( row ) )]
            num_row_lines = max( len( cell_line ) for cell_line in row_lines )
            
            if row_index <= 2:
                a = ansi.fore( 192 - 16, 192 - 16, 192 - 16 )
                c = ansi.back( 192, 192, 192 )
            elif row_index % 2 == 0:
                a = ansi.fore( 230 - 16, 255 - 16, 230 - 16 )
                c = ansi.back( 230, 255, 230 )
            else:
                a = ansi.fore( 255 - 16, 255 - 16, 255 - 16 )
                c = ansi.back( 255, 255, 255 )
            
            for row_line in range( num_row_lines ):
                if needs_l:
                    r.append( "\n" )
                else:
                    needs_l = True
                
                r.append( a + c + box.vr + box.em + ansi.RESET )
                
                for col, cell_lines in enumerate( row_lines ):
                    cell_line = cell_lines[row_line] if row_line < len( cell_lines ) else ""
                    
                    if col != 0:
                        r.append( a + c + box.em + box.vr + box.em + ansi.RESET + c )
                    else:
                        r.append( c )
                    
                    r.append( str( AnsiStr( cell_line ).ljust( col_lens[col] ) ) )
                    
                    r.append( ansi.RESET )
                
                r.append( a + c + box.em + box.vr + ansi.RESET )
        
        text = "".join( str( x ) for x in r )
        return text
    
    
    def on_format_hr( self ):
        return ansi.FORE_WHITE + ansi.BACK_BLUE + "-" * 80 + ansi.RESET
    
    
    def on_format_italic( self, x ):
        return ansi.ITALIC + x + ansi.ITALIC_OFF
    
    
    def on_format_key( self, x ):
        return ansi.FORE_GREEN + x + ansi.FORE_RESET
    
    
    def on_format_value( self, x ):
        return ansi.FORE_CYAN + x + ansi.FORE_RESET
    
    
    def on_format_error( self, x ):
        return self.format( ansi.BACK_RED + ansi.FORE_BRIGHT_WHITE, ansi.DIM + "ERROR: " + ansi.DIM_OFF + x )
    
    
    def on_format_code( self, x ):
        return ansi.FORE_BRIGHT_BLUE + x + ansi.FORE_RESET
    
    
    def on_format_file( self, x ):
        return ansi.FORE_YELLOW + x + ansi.FORE_RESET
    
    
    def on_format_warning( self, x ):
        return self.format( ansi.BACK_BRIGHT_YELLOW + ansi.FORE_BLACK, ansi.DIM + "WARNING: " + ansi.DIM_OFF + x )
    
    
    def on_format_positive( self, x ):
        return self.format( ansi.FORE_GREEN, x )
    
    
    def on_format_negative( self, x ):
        return self.format( ansi.FORE_RED, x )
    
    
    def on_format_neutral( self, x ):
        return self.format( ansi.FORE_CYAN, x )
    
    
    def on_format_command( self, x ):
        # Underline slows down iTerm2 massively
        # return self.format( ansi.UNDERLINE, x )
        return self.format( ansi.BOLD, x )
    
    
    def format( self, prefix, x ):
        return prefix + x.replace( self.RESET, ansi.RESET + prefix ) + self.RESET
    
    
    def on_format_verbose( self, x ):
        return self.format( ansi.DIM, x )
    
    
    def on_format_list( self, rows, ordered ):
        r = []
        
        for i, x in enumerate( rows ):
            if ordered:
                r.append( "{}{}.{} {}".format( ansi.DIM, str( i + 1 ), ansi.DIM_OFF, x ) )
            else:
                r.append( "{}•{} {}".format( ansi.DIM, ansi.DIM_OFF, x ) )
        
        return "\n".join( r )



import sys
import time
import hashlib
from typing import List

from mhelper import ansi
from sxsxml import SxsAnsiWriter, SxsHtmlWriter, SxsWriter
import pyperclip

from sxsxml_tests.case_parser import get_cases


class _TestFailedError( Exception ):
    pass


def dprint( x, end = "\n" ):
    import sys
    
    sys.__stdout__.write( x )
    if end:
        sys.__stdout__.write( end )


def __check_result( *,
                    name,
                    command_text,
                    command_hash,
                    writer: SxsWriter,
                    delay,
                    checker: List[str] = None ):
    name = name.upper()
    qq = ansi.BACK_LIGHT_BLACK + ansi.FORE_BLACK
    qe = ansi.RESET
    
    dprint( ansi.FORE_YELLOW + "§ " + ansi.RESET )
    dprint( ansi.FORE_YELLOW + "§ **************************************** {} TEST BEGINS ****************************************".format( name ) + ansi.RESET )
    dprint( ansi.FORE_YELLOW + "§" + ansi.RESET )
    
    checker.clear()
    
    __run_test( text = command_text,
                delay = delay,
                writer = writer )
    
    dprint( ansi.FORE_YELLOW + "§" + ansi.RESET )
    dprint( ansi.FORE_YELLOW + "§ INPUT" + ansi.RESET )
    dprint( ansi.FORE_YELLOW + "§" + ansi.RESET )
    for i, x in enumerate( command_text ):
        dprint( "{}RX{:03}[{}{}{}]{}".format( qq, i, qe, __fmt_easy( x ), qq, qe ) )
    
    dprint( ansi.FORE_YELLOW + "§" + ansi.RESET )
    dprint( ansi.FORE_YELLOW + "§ OUTPUT" + ansi.RESET )
    dprint( ansi.FORE_YELLOW + "§" + ansi.RESET )
    
    for i, x in enumerate( checker ):
        dprint( "{}TX{:03}[{}{}{}]{}".format( qq, i, qe, __fmt_easy( x ), qq, qe ) )
    
    dprint( ansi.FORE_YELLOW + "§" + ansi.RESET )
    dprint( ansi.FORE_YELLOW + "§ APPEARANCE" + ansi.RESET )
    dprint( ansi.FORE_YELLOW + "§" + ansi.RESET )
    content = "".join( checker )
    dprint( "{}CNTNT[{}{}{}]{}".format( qq, qe, content, qq, qe ) )
    pyperclip.copy( content )
    actual = hashlib.sha1( content.encode( 'utf-8' ) ).hexdigest()
    
    if command_hash and command_hash != actual:
        dprint( ansi.FORE_YELLOW + "§ " + ansi.RESET )
        dprint( ansi.FORE_YELLOW + "§ *********************************************************************************************" + ansi.RESET )
        dprint( ansi.FORE_YELLOW + "§ **************************************** TEST FAILED ****************************************" + ansi.RESET )
        dprint( ansi.FORE_YELLOW + "§ *********************************************************************************************" + ansi.RESET )
        dprint( ansi.FORE_YELLOW + "§ TEST: {}".format( name ) + ansi.RESET )
        dprint( ansi.FORE_YELLOW + "§ The actual hash '{}' does not match the expected hash '{}'.".format( actual, command_hash ) + ansi.RESET )
        dprint( ansi.FORE_YELLOW + "§ " + ansi.RESET )
        raise _TestFailedError( "Test failed." )
    
    dprint( ansi.FORE_YELLOW + "§ " + ansi.RESET )
    dprint( ansi.FORE_YELLOW + "§ ********** {} TEST {} **********".format( name, "PASSED" if command_hash else "UNCHECKED" ) + ansi.RESET )
    dprint( ansi.FORE_YELLOW + "§" + ansi.RESET )


def __fmt_easy( x ):
    qq = ansi.FORE_WHITE
    qe = ansi.RESET
    x = x.replace( "\033", ansi.BACK_MAGENTA + qq + "ESC" + qe )
    x = x.replace( "\n", ansi.BACK_BLUE + qq + "LF" + qe )
    x = x.replace( "\r", ansi.BACK_RED + qq + "CR" + qe )
    x = x.replace( " ", "·" )
    return x


def __run_test( *,
                text: List[str],
                delay: float,
                writer: SxsWriter ):
    delay /= 100
    
    for line in text:
        if line.startswith( "!REPEAT " ):
            s, e, line = line[len( "!REPEAT " ):].split( " ", 2 )
            for n in range( int( s ), int( e ) ):
                writer.write( line.format( n ) + "\n" )
                time.sleep( delay )
        else:
            writer.write( line + "\n" )


# noinspection SpellCheckingInspection
def run_tests( *,
               test_name: str = None,
               writer: SxsWriter,
               delay: float = 1,
               _writer_name: str = "none",
               _checker: List[str] = None ):
    commands = get_cases( _writer_name )
    
    if test_name is not None:
        commands = [x for x in commands if x[0] == test_name]
    
    for name, hash, text in commands:
        if _checker is not None:
            __check_result( name = _writer_name + " / " + name,
                            command_text = text,
                            command_hash = hash,
                            writer = writer,
                            delay = delay,
                            checker = _checker )
        else:
            __run_test( text = text,
                        delay = delay,
                        writer = writer )
    
    if _checker is not None:
        print( "+" )
        print( "+ **************************************************************************************************" )
        print( "+ **************************************** ALL TESTS PASSED ****************************************" )
        print( "+ **************************************************************************************************" )
        print( "+" )


def main():
    gui = False
    delay = 0
    echo = False
    
    for arg in sys.argv[1:]:
        if arg == "gui":
            gui = True
        elif arg == "delay":
            delay = 1
            echo = True
    
    if gui:
        from sxsxml_tests import gui
        gui.launch()
        return
    
    checker = []
    
    if echo:
        def ec( x ):
            checker.append( x )
            sys.__stderr__.write( x )
        
        
        tgt = ec
    else:
        tgt = checker.append
    
    try:
        run_tests( delay = delay,
                   _checker = checker,
                   writer = SxsAnsiWriter( tgt ),
                   _writer_name = "ansi" )
        
        run_tests( delay = delay,
                   _checker = checker,
                   writer = SxsHtmlWriter( tgt ),
                   _writer_name = "html" )
        
        run_tests( delay = delay,
                   _checker = checker,
                   writer = SxsHtmlWriter( tgt,
                                           scripted = False ),
                   _writer_name = "htms" )
    except _TestFailedError:
        exit( 1 )


if __name__ == "__main__":
    main()



cases = """
******** BASIC TEXT *****
** ansi = 09fac8dbfd27bd9b4d23a00eb648aa751789536d
** html = b0e305a47932b35656ee73dce111b273b529eb51
** htms = b0e305a47932b35656ee73dce111b273b529eb51 
Hello, world!


******** BASIC FORMATTING *****
** ansi = 4f7e480e94d30caa4eafc519c44f7a6ead26a07c
** html = 266c6ef1819c5a96e671e622b80682868ab2fc03
** htms = 266c6ef1819c5a96e671e622b80682868ab2fc03
one two <key>three</key> four <value>five</value> eight
one two <key>three</key> four <value>five</value> eight
one two <key>three</key> four <value>five</value> eight
the file <file>/alpha/beta/gamma.txt</file> is for this code
<warning>This is a warning</warning>
<error>This is an error</error>


******** TABLE *****
** ansi = 9b3d11bd85ece67d2b1fadcd2bdc4f1ec6ba6885
** html = 23577b63efd885a4882f962df4dca81b73615f5c
** htms = 23577b63efd885a4882f962df4dca81b73615f5c
<table>
<tr><td>one</td><td>two</td><td>three</td></tr>
<tr><td>four</td><td>five</td><td>six</td></tr>
<tr><td>seven</td><td>eight</td><td>nine</td></tr>
</table>


******** SECTIONS *****
** ansi = b90bd800bc4ef71c6b01ad252fa96f616d52f2bf
** html = 9e566f04556f13c5f7147ed8c77c1dc07efb2d90
** htms = 9e566f04556f13c5f7147ed8c77c1dc07efb2d90
section 0 text
section 0 text
<section name="section 1">
section 1 text
section 1 text
<section name="section 2">
section 2 text
section 2 text
<section name="section 3">
section 3 text
section 3 text
<section name="section 4">
section 4 text
section 4 text
<section name="section 5">
section 5 text
section 5 text
<section name="section 6">
section 6 text
section 6 text
</section>
section 5 text
section 5 text
</section>
section 4 text
section 4 text</section>section 3 text
section 3 text
</section>
section 2 text
section 2 text
</section>
section 1 text
section 1 text
</section>
section 0 text
section 0 text


******** HORIZONTAL RULE *****
** ansi = 86e093a4bf1dbc8ec99c5be89228c4c107355ced
** html = 2a57db4a8a4fb1497452cee90c9b2dc3862381d2
** htms = 2a57db4a8a4fb1497452cee90c9b2dc3862381d2
before rule
<hr></hr>
after rule


******** HORIZONTAL RULE *****
** ansi = 86e093a4bf1dbc8ec99c5be89228c4c107355ced
** html = 2a57db4a8a4fb1497452cee90c9b2dc3862381d2
** htms = 2a57db4a8a4fb1497452cee90c9b2dc3862381d2
before rule
<hr/>
after rule


******** PROGRESS BARS (BASIC) *****
** ansi = 494b20bb9503abdb15c8df5ee1f05adb1c2c0d66
** html = 2fdb4eb3855e675d63980be378f7fa957caf53c0
** htms = 2fdb4eb3855e675d63980be378f7fa957caf53c0
<action name="MASTER" max="3">
<progress value="0"></progress>
<action name="ACT1" max="100">
!REPEAT 0 100 <progress value="{}">progress text</progress>
</action>
<progress value="1"></progress>
<action name="ACT2" max="100">
!REPEAT 0 100 <progress value="{}">progress text</progress>
</action>
<progress value="2"></progress>
<action name="ACT3" max="100">
!REPEAT 0 100 <progress value="{}">progress text</progress>
</action>
</action>


******** PROGRESS BARS (WITH INTERRUPTIONS) *****
** ansi = 0a3f3b3146ceed07d90e13e3cfeaa5bf8d613250
** html = 03730090c81f84af037b2eacf12d91e40913d521
** htms = 03730090c81f84af037b2eacf12d91e40913d521
<action name="My action master" max="3">
This text at the start of the master.
This text at the start of the master.
<progress value="1"></progress>
<action name="My action 1" max="100">
This text at the start of the action.
This text at the start of the action.
!REPEAT 0 50 <progress value="{}"></progress>
* Hello, I'm interrupting things!
* REPEAT 51 100 <progress value="{}"></progress>
This text at the end of the action.
This text at the end of the action.
</action>
<progress value="2"></progress>
<action name="My action 2" max="100">
This text at the start of the action.
This text at the start of the action.
!REPEAT 0 50 <progress value="{}"></progress>
* Hello, I'm interrupting things!
#REPEAT 51 100 <progress value="{}"></progress>
This text at the end of the action.
This text at the end of the action.
</action>
<progress value="3"></progress>
<action name="My action 3" max="100">
This text at the start of the action.
This text at the start of the action.
!REPEAT 0 50 <progress value="{}"></progress>
* Hello, I'm interrupting things!
!REPEAT 51 100 <progress value="{}"></progress>
This text at the end of the action.
This text at the end of the action.
</action>
This text at the end of the master.
This text at the end of the master.
</action>


******** QUESTION *****
** ansi = f352554bbe0c928330ad22cdb46fd73e1d3640b9
** html = 0bbf188f14c874af25151f6c17cd19c3660f3803
<question>
This is the question text.
<option>Alpha</option>
<option>Beta</option>
<option>Gamma</option>
<option>Delta</option>
</question>
"""



from typing import Callable, Tuple, List, Dict


class _DirectedOutput:
    def __init__( self, target ):
        self.target = target
    
    
    def write( self, data ):
        self.target( data )
    
    
    def flush( self ):
        pass


class XsProgress:
    """
    Defines progress on an action.
    """
    
    
    def __init__( self, id: str, value: str, max: str, text: str, elapsed: str ):
        self.id: str = id
        
        try:
            self.value = int( value )
        except ValueError:
            self.value = 0
        
        try:
            self.max = int( max )
        except:
            self.value = 0
        
        self.text = text
        self.elapsed = elapsed
    
    
    @property
    def is_end( self ):
        return self.value == self.max
    
    
    @property
    def percent( self ):
        if self.is_end:
            return 100
        elif self.max:
            return int( (self.value / self.max) * 100 )
        else:
            return 0


class XsElement:
    """
    Represents an element of the form ``'<name attr1="value1" attr2="value2">'``
    """
    
    
    def __init__( self, raw: str, name: str, attr: Dict[str, str] ) -> None:
        self.raw = raw
        self.name = name
        self.attr = attr
    
    
    def get( self, item, default = "" ):
        return self.attr.get( item, default )
    
    
    def __repr__( self ) -> str:
        return "XsElement({}, {})".format( repr( self.name ), repr( self.attr ) )
    
    
    def __str__( self ):
        return "<{} {}>".format( self.name, " ".join( "{} = {}".format( x, y ) for x, y in self.attr.items() ) )


class XsStack:
    """
    Element stacking
    """
    
    
    def __init__( self, target: Callable[[str], None] ):
        super().__init__()
        self.__target = target
        self.__stack: Tuple[XsElement, List[str]] = (None, [])
        self.__stacks: List[Tuple[XsElement, List[str]]] = [self.__stack]
    
    
    def write( self, x: str ) -> None:
        if len( self.__stacks ) == 1:
            self.__target( x )
            return
        
        self.__stack[1].append( x )
    
    
    def push( self, e: XsElement ):
        self.__stack = e, []
        self.__stacks.append( self.__stack )
    
    
    def pop( self, e: XsElement ):
        text = "".join( self.__stack[1] )
        a = self.__stacks.pop()[0]
        
        if a is not e:
            raise ValueError( "Tried to pop the element '{}' from the stack but element at the top of the stack is '{}' (after '{}')."
                              "Check the element was pushed and in the correct order.".format( e, a, ", ".join( str( x[0] ) for x in self.__stacks ) ) )
        
        self.__stack = self.__stacks[-1]
        return text
    
    
    def reset( self ):
        self.__stack = (None, [])
        self.__stacks = [self.__stack]



from distutils.core import setup


setup( name = "sxsxml",
       url = "https://bitbucket.org/mjr129/sxsxml",
       version = "0.0.0.5",
       description = "XML stream to stdout with ANSI or HTML formatting",
       author = "Martin Rusilowicz",
       license = "https://www.gnu.org/licenses/agpl-3.0.html",
       packages = ["sxsxml", "sxsxml.sxs", "sxsxml_tests"],
       install_requires = ["mhelper"],
       python_requires = ">=3.6"
       )



import sys
import time

from sxsxml.abstract import XsStackedWriter
from sxsxml.support import XsProgress


class SxsWriter( XsStackedWriter ):
    """
    ABSTRACT
    
    Base class for SXS formatted streams.
    
    See ``readme.rst`` for the expected markup.
    """
    _CELL_START = "<$CS>"
    _CELL_END = "<$CE>"
    _ROW_START = "<$RS>"
    _ROW_END = "<$RE>"
    _START_TIME_TAG = "<$ST>"
    _NEWLINE = "<$NL>"
    
    
    def __init__( self, out = sys.__stdout__.write ):
        super().__init__( out )
        
        self.__sections = []
        self.__special_text = False
        self.__ignore_next_eol = False
        self.__in_table = False
        self.__no_newlines = 0
        self.section_name = "Untitled main section"
        
        self.map( ["document"], self.__enter_document, self.__exit_document )
        self.map( ["action"], self.__enter_action, self.__exit_action )
        self.map( ["argument"], self._enter_default, self.__exit_argument )
        self.map( ["literal", "tt", "code"], self._enter_default, self.__exit_code )
        self.map( ["command"], self._enter_default, self.__exit_command )
        self.map( ["error"], self._enter_default, self.__exit_error )
        self.map( ["file"], self._enter_default, self.__exit_file )
        self.map( ["heading"], self.__enter_heading, self.__exit_heading )
        self.map( ["hr"], self.__enter_hr, self.__exit_hr )
        self.map( ["emphasis", "em", "i", "cite", "title_reference"], self._enter_default, self.__exit_i )
        self.map( ["key"], self._enter_default, self.exit_key )
        self.map( ["li"], self._enter_default, self.__exit_li )
        self.map( ["negative"], self._enter_default, self.__exit_negative )
        self.map( ["neutral"], self._enter_default, self.__exit_neutral )
        self.map( ["ol"], self._enter_default, self.__exit_ol )
        self.map( ["option"], self._enter_default, self.__exit_option )
        self.map( ["br", "p", "paragraph"], self._enter_default, self.__exit_paragraph )
        self.map( ["positive"], self._enter_default, self.__exit_positive )
        self.map( ["progress"], self.__enter_progress, self.__exit_progress )
        self.map( ["question"], self.__enter_question, self.__exit_question )
        self.map( ["section"], self.enter_section, self.__exit_section )
        self.map( ["stderr"], self._enter_default, self.__exit_stderr )
        self.map( ["stdout"], self._enter_default, self.__exit_stdout )
        self.map( ["strong", "b"], self._enter_default, self.__exit_i )
        self.map( ["system"], self.__enter_system, self.__exit_system )
        self.map( ["table"], self.__enter_table, self.__exit_table )
        self.map( ["td"], self._enter_default, self.__exit_td )
        self.map( ["title"], self._enter_default, self.__exit_title )
        self.map( ["tr"], self._enter_default, self.__exit_tr )
        self.map( ["definition_list", "dl", "bullet_list", "ul"], self._enter_default, self.__exit_ul )
        self.map( ["value"], self._enter_default, self.__exit_value )
        self.map( ["verbose"], self.__enter_verbose, self.__exit_verbose )
        self.map( ["warning"], self._enter_default, self.__exit_warning )
    
    
    def __consume_eol( self ):
        self.on_consume_eol()
        self.ignore_next_eol()
    
    
    def ignore_next_eol( self ):
        self.__ignore_next_eol = True
    
    
    @property
    def section_depth( self ):
        # This prevents an indent being carried over into a tables cells
        # TODO: Find a better way that allows headings in cells
        if self.__in_table:
            return 0
        
        return len( self.__sections )
    
    
    def __enter_document( self ):
        self.__no_newlines += 1
    
    
    def __exit_document( self ):
        self.__no_newlines -= 1
        assert self.__no_newlines >= 0
    
    
    def on_text( self, x: str ):
        if not x:
            return
        
        if self.__special_text:
            super().on_text( x )
            return
        
        # If we are consuming an EOL then consume it...
        if self.__ignore_next_eol:
            self.__ignore_next_eol = False
            
            if x.startswith( "\n" ):
                x = x[1:]
            
            if not x:
                return
        
        if self.__no_newlines:
            x = x.replace( "\n", " " )
        
        x = x.replace( self._NEWLINE, "\n" )
        
        r = self.on_format_text( x )
        
        if r:
            super().on_text( r )
    
    
    def on_format_text( self, x ):
        raise NotImplementedError( "abstract" )
    
    
    #
    # SECTION
    #
    def enter_section( self ):
        self.__sections.append( self.element )
        self.__print_title( self.element.get( "name" ) )
        self.__consume_eol()
    
    
    def __exit_title( self ) -> None:
        self.__print_title( self.pop() )
    
    
    def __print_title( self, name ):
        self.section_name = name
        
        if name:
            r = self.on_format_heading( len( self.__sections ), name )
            x = self.__sections.pop( -1 )
            self.on_text( r )
            self.__sections.append( x )
    
    
    def on_format_heading( self, l, x ) -> str:
        raise NotImplementedError( "abstract" )
    
    
    def __exit_section( self ) -> None:
        assert self.__sections.pop() is self.element
        r = self.on_format_section_end()
        if r:
            self.on_text( r )
        self.__consume_eol()
    
    
    def on_consume_eol( self ):
        raise NotImplementedError( "abstract" )
    
    
    def __enter_heading( self ):
        self.push()
    
    
    def __exit_heading( self ):
        self.on_text( self.on_format_heading( int( self.element.get( "level", 1 ) ), self.pop() ) + self.on_format_section_end() )
    
    
    #
    # SYSTEM
    #
    def __enter_system( self ):
        self.element.attr["name"] = "System"
        self.enter_section()
    
    
    def __exit_system( self ):
        self.__exit_section()
    
    
    #
    # GENERAL
    #
    
    def on_format_positive( self, x ):
        raise NotImplementedError( "abstract" )
    
    
    def on_format_negative( self, x ):
        raise NotImplementedError( "abstract" )
    
    
    def on_format_neutral( self, x ):
        raise NotImplementedError( "abstract" )
    
    
    #
    # PROGRESS
    #
    def __enter_action( self ):
        setattr( self.element, self._START_TIME_TAG, time.time() )
        self.enter_section()
    
    
    def __exit_action( self ):
        self.__print_progress( self.__sections[-1].get( "max", "?" ), "" )
        self.__exit_section()
    
    
    def __enter_progress( self ):
        self.push()
        self.__special_text = True
    
    
    def __exit_progress( self ):
        self.__special_text = False
        value = self.get( "value", "?" )
        self.__print_progress( value, self.pop() )
    
    
    def __print_progress( self, value, text ):
        start = getattr( self.__sections[-1], self._START_TIME_TAG )
        elapsed = time.time() - start
        max = self.__sections[-1].get( "max", "?" )
        text = text.replace( "\n", "" )
        p = XsProgress( self.section_name, value, max, text, elapsed )
        r = self.on_format_progress( p )
        super().on_text( r )
        self.on_formatted_progress( p )
    
    
    def on_formatted_progress( self, p: XsProgress ):
        raise NotImplementedError( "abstract" )
    
    
    def on_format_progress( self, p: XsProgress ):
        raise NotImplementedError( "abstract" )
    
    
    #
    # QUESTION
    #
    def __enter_question( self ):
        self.push()
    
    
    def __exit_question( self ):
        question, options = _ss_split( self.pop(), self._ROW_START, self._ROW_END )
        question = self.on_format_question( question.strip(), options )
        
        super().on_text( question )
    
    
    def on_format_question( self, question, options ):
        raise NotImplementedError( "abstract" )
    
    
    def __exit_option( self ):
        self.on_text( self._ROW_START + self.pop() + self._ROW_END )
        self.__consume_eol()
    
    
    #
    # TABLE
    #
    def __exit_td( self ) -> None:
        self.on_text( self._CELL_START + self.pop() + self._CELL_END )
    
    
    def __exit_tr( self ) -> None:
        self.on_text( self._ROW_START + self.pop() + self._ROW_END )
    
    
    def __enter_table( self ):
        self.__in_table = True
        self.push()
    
    
    def __exit_table( self ) -> None:
        self.__in_table = False
        rows = _ss_split( self.pop(), self._ROW_START, self._ROW_END )[1]
        rows = [_ss_split( y, self._CELL_START, self._CELL_END )[1] for y in rows]
        cols = max( len( row ) for row in rows )
        rows = [row + ([""] * (cols - len( row ))) for row in rows]
        text = self.on_format_table( rows )
        
        self.on_text( text )
    
    
    def on_format_table( self, rows ):
        raise NotImplementedError( "abstract" )
    
    
    #
    # LISTS
    #
    def __exit_li( self ):
        self.on_text( self._ROW_START + self.pop() + self._ROW_END )
    
    
    def __exit_ul( self ):
        rows = _ss_split( self.pop(), self._ROW_START, self._ROW_END )[1]
        text = self.on_format_list( rows, ordered = False )
        self.on_text( text )
    
    
    def __exit_ol( self ):
        rows = _ss_split( self.pop(), self._ROW_START, self._ROW_END )[1]
        text = self.on_format_list( rows, ordered = True )
        self.on_text( text )
    
    
    #
    # VERBOSE
    #
    def __enter_verbose( self ):
        self.push()
    
    
    def __exit_verbose( self ):
        self.on_text( self.on_format_verbose( self.pop() ) )
    
    
    #
    # BASIC FORMATTING
    #
    def __enter_hr( self ) -> None:
        pass
    
    
    def __exit_hr( self ) -> None:
        self.on_text( self.on_format_hr() )
        self.__consume_eol()
    
    
    def on_format_hr( self ):
        raise NotImplementedError( "abstract" )
    
    
    def __exit_i( self ) -> None:
        self.on_text( self.on_format_italic( self.pop() ) )
    
    
    def exit_key( self ) -> None:
        self.on_text( self.on_format_key( self.pop() ) )
    
    
    def on_format_italic( self, x ):
        raise NotImplementedError( "abstract" )
    
    
    def on_format_key( self, x ):
        raise NotImplementedError( "abstract" )
    
    
    def __exit_value( self ) -> None:
        self.on_text( self.on_format_value( self.pop() ) )
    
    
    def on_format_value( self, x ):
        raise NotImplementedError( "abstract" )
    
    
    def __exit_file( self ) -> None:
        self.on_text( self.on_format_file( self.pop() ) )
    
    
    def on_format_file( self, x ):
        raise NotImplementedError( "abstract" )
    
    
    def __exit_code( self ) -> None:
        self.on_text( self.on_format_code( self.pop() ) )
    
    
    def on_format_code( self, x ):
        raise NotImplementedError( "abstract" )
    
    
    def __exit_error( self ) -> None:
        self.on_text( self.on_format_error( self.pop() ) )
    
    
    def on_format_error( self, x ):
        raise NotImplementedError( "abstract" )
    
    
    def __exit_paragraph( self ):
        self.on_text( self.pop() + self._NEWLINE )
    
    
    def __exit_positive( self ) -> None:
        self.on_text( self.on_format_positive( self.pop() ) )
    
    
    def __exit_negative( self ) -> None:
        self.on_text( self.on_format_negative( self.pop() ) )
    
    
    def __exit_neutral( self ) -> None:
        self.on_text( self.on_format_neutral( self.pop() ) )
    
    
    def __exit_warning( self ) -> None:
        self.on_text( self.on_format_warning( self.pop() ) )
    
    
    def __exit_command( self ) -> None:
        self.on_text( self.on_format_command( self.pop() ) )
    
    
    def __exit_stdout( self ) -> None:
        self.on_text( self.on_format_positive( self.pop() ) )
    
    
    def __exit_stderr( self ) -> None:
        self.on_text( self.on_format_negative( self.pop() ) )
    
    
    def __exit_argument( self ) -> None:
        self.on_text( self.on_format_key( self.pop() ) )
    
    
    def on_format_warning( self, x ):
        raise NotImplementedError( "abstract" )
    
    
    def on_format_command( self, x ):
        raise NotImplementedError( "abstract" )
    
    
    def on_format_section_end( self ):
        raise NotImplementedError( "abstract" )
    
    
    def on_format_verbose( self, x ):
        raise NotImplementedError( "abstract" )
    
    
    def on_format_list( self, rows, ordered ):
        raise NotImplementedError( "abstract" )


def _ss_split( text, prefix, suffix ):
    z = []
    v = []
    x = text.split( prefix )
    
    z.append( x[0] )
    
    for y in x[1:]:
        a, b = y.split( suffix, 1 )
        v.append( a )
        z.append( b )
    
    return "".join( z ), v



# noinspection PyPackageRequirements
from PyQt5.QtWebEngineWidgets import QWebEngineView

import sxsxml
import re
import sys
import mhelper_qt as qt

from sxsxml_tests import case_parser


class FrmMain( qt.QMainWindow ):
    def __init__( self ):
        super().__init__()
        
        self.lst_inputs = qt.QListWidget()
        self.lst_inputs.itemDoubleClicked.connect( self.itemDoubleClicked )
        self.fra_input = qt.QFrame()
        self.lay_input = qt.QVBoxLayout()
        self.txt_input = qt.QTextEdit()
        self.btn_input = qt.QPushButton()
        self.btn_input.setText( "Render" )
        self.btn_input.clicked.connect( self.clicked )
        self.cmb_input = qt.QComboBox()
        self.cmb_input.addItems( ["ansi", "html", "html+"] )
        self.txt_output = QWebEngineView()
        self.spt_main = qt.QSplitter()
        
        self.spt_main.addWidget( self.lst_inputs )
        self.spt_main.addWidget( self.fra_input )
        self.spt_main.addWidget( self.txt_output )
        
        self.fra_input.setLayout( self.lay_input )
        self.lay_input.addWidget( self.txt_input )
        self.lay_input.addWidget( self.cmb_input )
        self.lay_input.addWidget( self.btn_input )
        
        self.setCentralWidget( self.spt_main )
        
        self.cases = { }
        
        for name, hash, content in case_parser.get_cases( "" ):
            item = qt.QListWidgetItem()
            item.setText( name )
            self.lst_inputs.addItem( item )
            self.cases[name] = "\n".join( content )
    
    
    def itemDoubleClicked( self, item ):
        self.txt_input.setPlainText( self.cases[item.text()] )
    
    
    def clicked( self ):
        rx = re.compile("\033\[[0-9]+m")
        i = self.cmb_input.currentIndex()
        op = []
        
        if i == 0:
            o = sxsxml.SxsAnsiWriter( op.append )
        elif i == 1:
            o = sxsxml.SxsHtmlWriter( op.append, scripted = False )
        elif i == 2:
            o = sxsxml.SxsHtmlWriter( op.append )
        else:
            raise ValueError( "No selection" )
        
        o.write( self.txt_input.toPlainText() )
        
        txt = "\n".join( op )
        
        if i == 0:
            sys.__stderr__.write( txt )
            txt = rx.sub("␛", txt)
            txt = txt.replace( "<", "&lt;" )
            txt = txt.replace( ">", "&gt;" )
            txt = txt.replace( "\033", "␛" )
            txt = txt.replace( "\n", "↲<br/>" )
            txt = txt.replace( "\r", "↞<br/>" )
            txt = txt.replace( " ", "␠" )
            txt = "<code>{}</code>".format( txt )
        
        txt = "<html><head><style>{}</style></head><body>{}</body></html>".format( sxsxml.SxsHtmlWriter.BASIC_CSS, txt )
        
        self.txt_output.setHtml( txt )


def launch():
    application = qt.QApplication( sys.argv )
    window = FrmMain()
    window.show()
    application.exec_()



def get_cases( _writer_name ):
    from sxsxml_tests.cases import cases
    commands = []
    current = None
    for line in cases.split( "\n" ):
        if not line:
            continue
        elif line.startswith( "**" ):
            line = line[2:].strip()
            
            if "=" in line:
                nam, _, val = line.partition( "=" )
                if nam.strip() == _writer_name:
                    current[1] = val.strip()
            else:
                current = [line.strip( " *" ).lower(), "", []]
                commands.append( current )
        else:
            current[2].append( line )
    return commands


from docutils import core, io


x = core.publish_parts(
        source="""
Title
=====

aaa

Title
=====

aaa

Subtitle
--------

aaa

Subtitle
--------

This is *italic*, **bold** and `citation` and ``teletype``.

* item
* item
* more
* more
* item
        """,
        writer_name='xml')["whole"]

if '<document source="&lt;string&gt;">' in x:
    x = x.split('<document source="&lt;string&gt;">',1 )[1]
else:
    x = x.split('<body>',1 )[1]
    x = x.split('</body>',1 )[0]
    



print(x)


from mhelper import file_helper

import sys

from sxsxml.sxs.base import SxsWriter
from sxsxml.support import XsProgress


class SxsHtmlWriter( SxsWriter ):
    """
    Writes HTML from an SXS stream.
    """
    PROGRESS_PREFIX = "<progress "
    PROGRESS_SUFFIX = ">"
    BASIC_CSS = """
    span.verbose
    {
    color: #C0C0C0;
    }
    
    span.progress_l
{
    position: absolute;
    left: 0px;
    background: #2196F3; 
    height:100%;
}

span.progress_r
{
    position: absolute;
    right: 0px;
}

span.progress_c
{
    position: absolute;
    left: 0px;
    width:100%;
    height:100%;
    text-align: center;
}

div.progress
{
    position: relative;
    background: white;
    width: 100%;
    height:16px;
    border: 1px solid silver;
    border-radius: 4px;
}

div.section
{
}

div.question
{
    style:"background: #FFFFE0";
}

a.answer
{
}

span.key
{
    font-weight : bold;
}

span.value
{
    font-style: italic;
}

span.code
{
    font-family: monospace;
}

a.file
{
    font-family: monospace;
}

span.error
{
    background: red; 
    color: white;
}

span.warning
{
    background: yellow; 
    color: black
}

span.positive
{
    color: green;
}

span.negative
{
    color: red;
}

span.neutral
{
    color: brown;
}

a.command
{
    font-family: monospace;
}
"""
    
    
    def __init__( self, out = sys.__stdout__.write, scripted = True ):
        super().__init__( out = out )
        self.progress_bars = set()
        self.last_eol = False
        self.scripted = scripted
    
    
    def on_format_text( self, x ):
        x = x.replace( "\n", "<br/>\n" )
        self.last_eol = x.endswith( "\n" )
        
        return x
    
    
    def on_format_heading( self, level, name ):
        if name:
            return '<div class="section"><h{0}>{1}</h{0}>'.format( level, name )
        else:
            return '<div style="section">'
    
    
    def _exit_default( self ):
        self.on_text( "!!Unknown element '{}'!!{}".format( self.element.name, self.pop() ) )
    
    
    def on_format_section_end( self ):
        return "</div>\n"
    
    
    def on_consume_eol( self ):
        pass
    
    
    def on_format_progress( self, p: XsProgress ):
        self.ignore_next_eol()
        
        i = p.id.replace( " ", "_" )
        
        if self.scripted:
            if i in self.progress_bars:
                # Bar exists - update it
                r = self.__mk_progress_bar_update_script( i, p )
            else:
                # Create a new bar
                self.progress_bars.add( i )
                r = self.__mk_progress_bar( i, p )
        else:
            if i in self.progress_bars:
                # Bar exists - update it
                self.progress_bars.add( i )
                r = "{}0 {} {}{}".format( self.PROGRESS_PREFIX, 1 if p.is_end else 0, i, self.PROGRESS_SUFFIX ) + self.__mk_simple_progress_bar( p )
            else:
                # Create a new bar
                self.progress_bars.add( i )
                r = "{}1 {} {}{}".format( self.PROGRESS_PREFIX, 1 if p.is_end else 0, i, self.PROGRESS_SUFFIX ) + self.__mk_simple_progress_bar( p )
        
        if p.is_end:
            self.progress_bars.remove( i )
        
        return r
    
    
    def __mk_progress_bar_update_script( self, i, p ):
        return ('<script>\n'
                'document.getElementById("{0}_l").style.width="{1}%";\n'
                'document.getElementById("{0}_l").innerHTML="{2}";\n'
                'document.getElementById("{0}_r").innerHTML="{3}";\n'
                'document.getElementById("{0}_c").innerHTML="{4}";\n'
                '</script>\n').format( i, p.percent, p.value, p.max, self.__get_progress_text( p ) )
    
    
    def __mk_progress_bar( self, i, p ):
        return ('<div class="progress">\n'
                '<span id="{0}_l" style="width:{1}%" class="progress_l">\n'
                '{2}\n'
                '</span>\n'
                '<span id="{0}_r" class="progress_r">\n'
                '{3}\n'
                '</span>\n'
                '<span id="{0}_c" class="progress_c">\n'
                '{4}\n'
                '</span>\n'
                '</div>\n').format( i, p.percent, p.value, p.max, self.__get_progress_text( p ) )
    
    
    def __get_progress_text( self, p ):
        return self.section_name + ": " + ("done" if p.is_end else p.text)
    
    
    def __mk_simple_progress_bar( self, p ):
        return "<span class='progress'>{}/{} = {}%</span>".format( p.value, p.max, p.percent, self.__get_progress_text( p ) )
    
    
    def on_format_question( self, question, options ):
        if not options:
            options = ("*",)
        
        return (
            '<div class="question">'
            '<p>{}</p>'
            '<p>'
            '{}'
            '</p>'
            '</div>\n').format( question, " ".join( ('[<a class="answer" href="{0}">{0}</a>]'.format( x ) for x in options) ) )
    
    
    def on_format_table( self, rows ):
        return (
            '<table>'
            '{}'
            '</table>').format(
                "".join( "<tr>{}</tr>".format(
                        "".join( "<td>{}</td>".format( cell ) for cell in row ) ) for row in rows ) )
    
    
    def on_format_hr( self ):
        return "<hr/>"
    
    
    def on_format_italic( self, x ):
        return '<i>{}</i>'.format( x )
    
    
    def on_format_key( self, x ):
        return '<span class="key">{}</span>'.format( x )
    
    
    def on_format_value( self, x ):
        return '<span class="value">{}</span>'.format( x )
    
    
    def on_format_file( self, x ):
        return '<a class="file" href="{}">{}</a>'.format( x, file_helper.get_filename( x ) )
    
    
    def on_format_code( self, x ):
        return '<span class="code">{}</span>'.format( x )
    
    
    def on_format_error( self, x ):
        return '<span class="error">{}</span>'.format( x )
    
    
    def on_format_warning( self, x ):
        return '<span class="warning">{}</span>'.format( x )
    
    
    def on_formatted_progress( self, p: XsProgress ):
        pass
    
    
    def on_format_positive( self, x ):
        return '<span class="positive">{}</span>'.format( x )
    
    
    def on_format_negative( self, x ):
        return '<span class="negative">{}</span>'.format( x )
    
    
    def on_format_neutral( self, x ):
        return '<span class="neutral">{}</span>'.format( x )
    
    
    def on_format_command( self, x ):
        return '<a class="command" href="{0}">{0}</a>'.format( x )
    
    
    def on_format_verbose( self, x ):
        return '<span class="verbose">{}</span>'.format( x )
    
    
    def on_format_list( self, rows, ordered ):
        return "<{0}>{1}</{0}>".format( "ol" if ordered else "ul", "<li>{}</li>".format( x for x in rows ) )



import sys
from typing import cast
from html.parser import HTMLParser

from mhelper import ansi_format_helper, Logger
from sxsxml.support import XsElement, XsStack


LOG = Logger( "SXS", False )


class XsWriter:
    """
    ABSTRACT
    
    Base class for `XsParser` output targets.
    
    The derived class should implement the abstract functions.
    
    .. note::
    
        `XsNestedOutput` may provide a more usable class in certain circumstances.
    """
    
    
    def __init__( self ) -> None:
        super().__init__()
        self.__stack = []
        self.__parser = _InternalParser( self )
    
    
    def on_text( self, x: str ) -> None:
        """
        Handles text written to stream.
        """
        raise NotImplementedError( "abstract" )
    
    
    def on_enter( self, e: XsElement ) -> None:
        """
        Handles entering an element.
        """
        raise NotImplementedError( "abstract" )
    
    
    def on_exit( self, e: XsElement ) -> None:
        """
        Handles exiting an element.
        """
        raise NotImplementedError( "abstract" )
    
    
    def on_error( self, ex: Exception ) -> bool:
        """
        Called to process any parser error.
         
        :param ex:  Error 
        :return:    Returning a false-like value or `None` ignores the error.
        """
        cast( None, ex )
        return True
    
    
    def on_flush( self ) -> None:
        """
        Called to perform a flush.
        """
        pass
    
    
    def flush( self ):
        self.on_flush()
    
    
    def reset( self ):
        """
        Reset to initial state.
        """
        self.on_reset()
        assert not self.__stack, "Derived class did not call base on_reset in '{}'".format( self )
    
    
    def on_reset( self ) -> None:
        """
        Reset to initial state.
        Derived class should call back to base class.
        """
        self.__parser.reset()
        self.__stack = []
    
    
    def write( self, data ):
        try:
            self.__parser.feed( data )
        except Exception as ex:
            sys.__stderr__.write( "FULL TEXT BEGINS\n" )
            sys.__stderr__.write( data )
            sys.__stderr__.write( "\nFULL TEXT ENDS\n" )
            sys.__stderr__.write( ansi_format_helper.format_traceback( ex ) )
            sys.__stderr__.write( "\n" )
            sys.__stderr__.write( "FULL TEXT PRECEDES ERROR MESSAGE\n" )
            self.__parser.reset()
            self.on_text( "(PARSER ERROR: SEE STDERR FOR DETAILS)\n" )
    
    
    def print( self, text: str, *, end: str = "\n" ) -> None:
        self.__parser.feed( text )
        
        if end:
            self.__parser.feed( end )


class _InternalParser( HTMLParser ):
    def __init__( self, owner: XsWriter ):
        super().__init__()
        self.__stack = []
        self.__owner = owner
    
    
    def handle_starttag( self, tag, attrs ):
        e = XsElement( "", tag, dict( attrs ) )
        self.__stack.append( e )
        self.__owner.on_enter( e )
    
    
    def handle_endtag( self, tag ):
        e = self.__stack.pop( -1 )
        
        if e.name != tag:
            self.error( "Entering tag is '{}' but closing tag is '{}'".format( e.name, tag ) )
        
        self.__owner.on_exit( e )
    
    
    def handle_data( self, data ):
        self.__owner.on_text( data )
    
    
    def error( self, message ):
        try:
            raise ValueError( "Error from HTML parser: " + message )
        except ValueError as ex:
            if self.__owner.on_error( ex ):
                raise


class XsStackedWriter( XsWriter ):
    """
    ABSTRACT
    
    Finds methods `enter_` and `exit_` in the derived class and uses these to format the output.
    
    The enter methods can call `push` and the exit methods `pop` to retrieve element text. 
    """
    
    
    def __init__( self, out = sys.stdout.write ):
        super().__init__()
        
        self.__enter_map = { }
        self.__exit_map = { }
        self.__stack = XsStack( out )
        self.element: XsElement = None
    
    
    def map( self, tags, enter, exit ):
        for tag in tags:
            if enter is not None:
                self.__enter_map[tag] = enter
            
            if exit is not None:
                self.__exit_map[tag] = exit
    
    
    def on_reset( self ):
        super().on_reset()
        self.__stack.reset()
    
    
    def get( self, key, default = "" ) -> str:
        return self.element.get( key, default )
    
    
    def on_enter( self, e: XsElement ):
        self.element = e
        fn = self.__enter_map.get( e.name, self._enter_default )
        LOG( "Entering {}", e, fn )
        fn()
        self.element = None
    
    
    def on_exit( self, e: XsElement ):
        self.element = e
        fn = self.__exit_map.get( e.name, self._exit_default )
        LOG( "Exiting {} {}", e, fn )
        fn()
        self.element = None
    
    
    def push( self ):
        LOG( "Pushing {}", self.element )
        self.__stack.push( self.element )
    
    
    def pop( self ) -> str:
        LOG( "Popping {}", self.element )
        return self.__stack.pop( self.element )
    
    
    def _enter_default( self ):
        self.push()
    
    
    def _exit_default( self ):
        self.pop()
    
    
    def on_text( self, x: str ) -> None:
        self.__stack.write( x )
    
    
    def on_error( self, ex: Exception ):
        self.__stack.reset()
        return True
